//! Macros were created because functions started to get really complicated.
//! These macros should not be exported out of the library as they are
//! not safe to use since they do no type checking.



/// Handles the repetative bit packing for big endian.
macro_rules! pack_big_endian
{
   ($buffer: ident, $target_type: ty, $bytes: expr) =>
   {
      {
         let mut value: $target_type;

         // Make sure that there is enough data to read
         // the bytes for this type.
         assert!($buffer.len() >= $bytes);

         // Start with a value of zero and or it with
         // the bits shifted values from the buffer.
         value = 0;
         for i in 0..$bytes
         {
            value |= ($buffer[i] as $target_type) << ((($bytes - i) - 1) * 8);
         }

         // Return the determined value.
         value
      }
   }
}

/// Handles the repetative bit packing for little endian.
macro_rules! pack_little_endian
{
   ($buffer: ident, $target_type: ty, $bytes: expr) =>
   {
      {
         let mut value: $target_type;

         // Make sure that there is enough data to read
         // the bytes for this type.
         assert!($buffer.len() >= $bytes);

         // Start with a value of zero and or it with
         // the bits shifted values from the buffer.
         value = 0;
         for i in 0..$bytes
         {
            value |= ($buffer[($bytes - 1) - i] as $target_type) <<
                     ((($bytes - i) - 1) * 8);
         }

         // Return the determined value.
         value
      }
   }
}

/// Handles the repetative big endian unpacking.
macro_rules! unpack_big_endian
{
   ($value: expr, $bytes: expr) =>
   {
      {
         let mut buffer: Vec<u8>;

         // Create an array with enough space for this value
         // and then bit shift the value into a buffer of bytes.
         buffer = Vec::with_capacity($bytes);
         for i in 0..$bytes
         {
            buffer.push(($value >> (($bytes - i) - 1) * 8) as u8);
         }

         // Return the buffer of bytes that represent this value.
         buffer
      }
   }
}

/// Handles the repetative little endian unpacking.
macro_rules! unpack_little_endian
{
   ($value: expr, $bytes: expr) =>
   {
      {
         let mut buffer: Vec<u8>;

         // Create an array with enough space for this value
         // and then bit shift the value into a buffer of bytes.
         buffer = Vec::with_capacity($bytes);
         for i in 0..$bytes
         {
            buffer.push(($value >> (i * 8)) as u8);
         }

         // Return the buffer of bytes that represent this value.
         buffer
      }
   }
}

/// Handles swapping from big endian to little endian.
macro_rules! swap_big_to_little_endian
{
   ($value: expr, $target_type: ty, $bytes: expr) =>
   {
      {
         let buffer: Vec<u8>;

         // Convert the big endian value to bytes.
         buffer = unpack_big_endian!($value, $bytes);

         // Convert the buffer of bytes to a little endian value.
         pack_little_endian!(buffer, $target_type, $bytes)
      }
   }
}

/// Handles swapping from little endian to big endian.
macro_rules! swap_little_to_big_endian
{
   ($value: expr, $target_type: ty, $bytes: expr) =>
   {
      {
         let buffer: Vec<u8>;

         // Convert the little endian value to bytes.
         buffer = unpack_little_endian!($value, $bytes);

         // Convert the buffer of bytes to a big endian value.
         pack_big_endian!(buffer, $target_type, $bytes)
      }
   }
}
